from typing import List
from typing import Tuple

import psycopg2 as pg2
from sbcommons.logging.lambda_logger import get_logger

logger = get_logger(__name__)


class RedshiftClient:
    """
    Class to interact with redshift DB.
    """

    def __init__(self, host: str, port: int, db_name: str, user: str, password: str):
        self.host = host
        self.port = port
        self.db_name = db_name
        self.user = user
        self.password = password

    def __enter__(self):
        self.connection = pg2.connect(
            host=self.host,
            port=self.port,
            dbname=self.db_name,
            user=self.user,
            password=self.password
        )
        self.cursor = self.connection.cursor()
        return self

    def __exit__(self, *_):
        self.cursor.close()
        self.connection.close()

    def kill_locks(self, table_name: str, exlude_users: List[str] = None):
        """
        Method to call to kill all acquired locks on a given DB table.

        :param table_name: name of the table to kill all locks for
        :param exlude_users: list of usernames for whom processes should not be terminated
        """
        exclude_users_cond = f"""    AND s.usename NOT IN ('{"', '".join(exlude_users)}');""" \
            if exlude_users else ';'

        locks_query = f'SELECT DISTINCT l.pid, s.usename ' \
            f'FROM pg_locks l ' \
            f'INNER JOIN pg_stat_all_tables t ON l.relation=t.relid ' \
            f'INNER JOIN pg_stat_activity s ON l.pid = s.procpid ' \
            f'WHERE t.schemaname || \'.\' || t.relname = \'{table_name}\' ' \
            f'    AND l."granted"' \
            f'{exclude_users_cond}'
        self.cursor.execute(locks_query)
        rows = self.cursor.fetchall()

        if rows:
            kill_lock_queries = [f'SELECT pg_terminate_backend({row[0]});' for row in rows]
            kill_lock_query = '\n'.join(kill_lock_queries)
            try:
                self.cursor.execute(kill_lock_query)
            except Exception as e:
                logger.warning(f'Exception caught when killing query! One of the PIDs in: '
                               f'{rows} (first column) was probably already terminated.')
                raise e

    def load_data(self, table, s3_bucket, s3_manifest_key):
        """
        Load data into event schema.

        :param table: table name to load data to
        :param s3_bucket: bucket where data exists
        :param s3_manifest_key: key to manifest file holding entries to load
        """

        staging_table = self._create_staging_table(table)
        self.copy_to_redshift(table_name=staging_table,
                              bucket=s3_bucket,
                              s3_key=s3_manifest_key,
                              options=['CSV', 'DELIMITER AS \'|\'', 'EMPTYASNULL', 'GZIP',
                                       'MANIFEST'],
                              should_commit=False)
        nr_events = self._count_rows(staging_table)
        rows_before = self._count_rows(table)
        self._deduplicate_and_insert(table, staging_table)
        rows_after = self._count_rows(table)
        rows_inserted = rows_after - rows_before
        self.drop_table(table=staging_table)
        logger.info(f'Inserted [{rows_inserted}] new rows into: [{table}] from {nr_events} events.')
        self.connection.commit()

    def unload_query_to_s3(self, bucket: str, path: str, unload_query: str,
                           options: List[str] = None):
        """
        This function takes a redshift query and outputs it to a specified s3 path.
        The options List is an additional unload arguments.
        You can see all arguments here https://docs.aws.amazon.com/redshift/latest/dg/r_UNLOAD.html
        """
        if not options:
            options = ["FORMAT AS CSV", "DELIMITER AS '|'", "NULL AS ''"]
        execute_query = f'''UNLOAD ('{unload_query}') 
                            TO 's3://{bucket}/{path}' 
                            iam_role 'arn:aws:iam::485206745971:role/RedshiftRole'
                            {' '.join(options)};
                            '''
        self.cursor.execute(execute_query)
        self.connection.commit()

    def vacuum_analyze(self, table: str):
        """
        Run vacuum & analyze on all tables in the provided schemas.

        :param table: str containing table to operate on
        """
        # VACUUM cannot run inside a transaction block
        self.connection.autocommit = True
        self.cursor.execute(f'VACUUM REINDEX {table};')

    def get_state(self, state_table: str, key: str):
        query = f'SELECT s.last_updated ' \
            f'FROM {state_table} s ' \
            f'WHERE s.key = \'{key}\';'
        self.cursor.execute(query)
        state = self.cursor.fetchone()
        return state[0] if state else None

    def upsert_state(self, state_table: str, key: str, value: str):
        self.kill_locks(state_table)

        if self.get_state(state_table, key):
            query = f'UPDATE {state_table} ' \
                f'SET last_updated = \'{value}\' ' \
                f'WHERE key = \'{key}\';'
        else:
            query = f'INSERT INTO {state_table} VALUES (\'{key}\', TIMESTAMP \'{value}\');'

        self.cursor.execute(query)
        self.connection.commit()

    def _create_staging_table(self, table):
        staging_table = f'{table.split(".")[1]}_staging'
        self.cursor.execute(f'CREATE TABLE {staging_table} (LIKE {table});')
        return staging_table

    def copy_to_redshift(self, table_name: str, bucket: str, s3_key: str,
                         options: List[str] = None,
                         columns: List[str] = None,
                         should_commit: bool = False):
        """
            This us a simple copy command native to redshift.
            Please note that the table_name must include the schema dot table name.
            for example view.funnel
            OBS! Set should_commit = True when this function is called to commit after run.
        """
        col_opts = f'''({', '.join(columns)})''' if columns else ''
        if not options:
            options = ['CSV', 'DELIMITER AS \'|\'', 'EMPTYASNULL', 'GZIP', 'MANIFEST']
        query = f'''
        COPY {table_name}{col_opts}
        FROM 's3://{bucket}/{s3_key}'
        CREDENTIALS 'aws_iam_role=arn:aws:iam::485206745971:role/RedshiftRole'
        {' '.join(options)};
        '''
        self.cursor.execute(query)
        if should_commit:
            self.connection.commit()

    def truncate_table(self, table_name: str):
        """
            OBS The TRUNCATE command commits the transaction in which it is run; therefore,
            you can't roll back a TRUNCATE operation, and a TRUNCATE command may commit other
            operations when it commits itself.
        """
        query = f'TRUNCATE {table_name};'
        self.cursor.execute(query)

    def _deduplicate_and_insert(self, table, staging_table):
        self.cursor.execute(f'CREATE TEMP TABLE {staging_table}_tmp (LIKE {staging_table});')
        self.cursor.execute(f'ALTER TABLE {staging_table}_tmp ADD COLUMN rank INT;')
        query = f'''
        INSERT INTO {staging_table}_tmp
        WITH deduplicated_keys AS (
            SELECT DISTINCT stg.key
            FROM {staging_table} stg
            EXCEPT
            SELECT t.key
            FROM {table} t
        ),
        one_row_per_key AS (
            SELECT
                stg.*,
                ROW_NUMBER() OVER (PARTITION BY stg.key ORDER BY stg.sent_at DESC) AS rank
            FROM deduplicated_keys dk
            INNER JOIN {staging_table} stg ON stg.key = dk.key
        )
        SELECT orpk.*
        FROM one_row_per_key orpk
        WHERE orpk.rank = 1;
        '''
        self.cursor.execute(query)
        self.cursor.execute(f'ALTER TABLE {staging_table}_tmp DROP COLUMN rank;')
        self.cursor.execute(f'INSERT INTO {table} SELECT * FROM {staging_table}_tmp;')
        self.drop_table(table=f'{staging_table}_tmp')

    def drop_table(self, table, should_commit: bool = False):
        self.cursor.execute(f'DROP TABLE {table};')
        if should_commit:
            self.connection.commit()

    def _count_rows(self, table: str) -> int:
        self.cursor.execute(f'SELECT COUNT(*) FROM {table};')
        nr_rows = self.cursor.fetchone()
        return nr_rows[0]

    def tables_in_schemas(self, schemas: List[str]) -> List[str]:
        query = f'''
        SELECT DISTINCT t.table_schema || '.' || t.table_name::TEXT AS table
        FROM information_schema.tables t
        WHERE t.table_schema IN ('{"', '".join(schemas)}')
        ORDER BY "table";
        '''
        self.cursor.execute(query)
        return [row[0] for row in self.cursor.fetchall()]

    def execute_query(self, query: str) -> Tuple[List[str], List[Tuple]]:
        """ Gets columns and rows for specified <query>.

        Args:
            query: The select statement to run.

        Returns:
             A tuple with two lists: i) A list of the column names and ii) A list of tuples
                representing the rows returned by the select query.
        """
        self.cursor.execute(query)
        return [d[0] for d in self.cursor.description], self.cursor.fetchall()



import functools
import traceback
from typing import Any

import boto3


def sns_client(region: str) -> Any:
    return boto3.client('sns', region_name=region)


def publish(topic_arn: str, region: str, message: str):
    client = sns_client(region)
    client.publish(
        TargetArn=topic_arn,
        Message=message
    )


def publish_to_sns_on_failure(topic_arn: str, region: str = 'eu-north-1'):
    """ Parametrizes a decorator function for publishing to an SNS topic on failure.

    If an exception is raised by the function we use the decorator on, the exception stack traceback
    is pushed as a message to the SNS topic defined by <topic_arn> and <region>. Then the caught
    exception is re-raised.

    Example use with lambda function:
        @publish_to_sns_on_failure(<arn_topic>, 'eu-north-1')
        def handler(event, context):
            ...
            <code that can raise an exception>
            ...
            return {'statusCode': 200}
            
    Args:
        topic_arn: AWS SNS topic to publish a message to.
        region: The AWS region of the SNS topic.
        
    Return:
        The actual decorator function for publishing to a specific SNS topic.
    """
    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                exc_traceback = traceback.format_exc()
                publish(topic_arn=topic_arn, region=region, message=exc_traceback)
                raise e
        return wrapper
    return actual_decorator



""" Class definition for a Klaviyo metric """
from abc import ABC
import logging
from string import Template
from sbcommons.crm.klaviyo.url import (get_filter_url,get_url_since_timestamp_filter_args,
                                       get_url_until_timestamp_filter_args,
                                       get_next_page_cursor_arg)
from urllib.parse import quote


class KlaviyoMetricError(Exception):
    """ Klaviyo metric exception class"""

    def __init__(self, msg):
        super(KlaviyoMetricError, self).__init__(msg)


class KlaviyoMetric(ABC):
    """ Base class for a Klaviyo metric

    Attributes:
        metric_name: Name of the Klaviyo metric.
    """
    # Metric related URLs and URL-templates used to make Klaviyo API calls
    _METRICS_INFO_URL = 'https://a.klaviyo.com/api/metrics/'
    _METRIC_EVENTS_BASE_URL = 'https://a.klaviyo.com/api/events/'
    _FILTER_TEMPLATE = Template('filter=$condition($metric_id,$metric_value)')
    _INCLUDE_TEMPLATE = Template('include=$value')

    def __init__(self, metric_name: str):
        self.metric_name = metric_name
        self._logger = logging.getLogger(__name__)

    @classmethod
    @property
    def include_template(cls):
        return cls._INCLUDE_TEMPLATE

    @classmethod
    @property
    def metrics_info_url(cls):
        return cls._METRICS_INFO_URL

    @classmethod
    @property
    def metric_events_url(cls):
        return cls._METRICS_INFO_URL

    @classmethod
    def get_metric_events_url(cls, metric_id: str):
        """ Constructs URL for getting metric events given a metric identifier. """
        return cls._METRIC_EVENTS_BASE_URL + f"?filter=" + get_filter_url(condition='equals',
                                                                          metric_id='metric_id',
                                                                          metric_value=
                                                                          f'"{metric_id}"')


    @classmethod
    def get_metrics_info_url_args(cls, page_cursor: str):
        return get_next_page_cursor_arg(page_cursor)




from typing import Any
from typing import Dict
from typing import Iterable
from typing import List

from sbcommons.crm.parser import CrmParser
from sbcommons.crm.parser import TranslationEntry


class KlaviyoParser(CrmParser):
    def __init__(self, translation_entries: List[TranslationEntry]):
        CrmParser.__init__(self, translation_entries)

    def parse(self, header: List[str], rows: List[Iterable[Any]]) -> List[Dict]:
        """
        list_data: A list of dictionaries where each dictionary corresponds to the customer
                profile we want to add to the list. The profile must have an identifier such as
                a mobile phone number or e-mail address to be added to the list.
        """
        return [self._parse_row(header, row) for row in rows]

    def _parse_row(self, header: List[str], row: Iterable[Any]) -> Dict[str, str]:
        f = {
            self.translation_dict[db_column_name].crm_attr_name: self._sanitize(db_column_name, val)
            for db_column_name, val in zip(header, row)
        }
        return f



from string import Template
from  urllib.parse import quote

_FILTER_TEMPLATE = Template('$condition($metric_id,$metric_value)')
_PAGE_CURSOR_ARG_TEMPLATE = Template('?page[cursor]=$page_cursor')
_SORT_ARG_TEMPLATE = Template('sort=$sort_criterion')
_FIELD_ARG_TEMPLATE = Template('fields[$metric]=$value')


def get_field_arg(metric, value):
    return _FIELD_ARG_TEMPLATE.substitute(metric=metric,
                                          value=value)


def get_filter_url(condition, metric_id, metric_value):
    return _FILTER_TEMPLATE.substitute(condition=condition,
                                       metric_id=metric_id,
                                       metric_value=quote(metric_value))


def get_sort_arg(sort_criterion):
    """ Gets the argument string for the Klaviyo API call given the parameters

            Args:
                sort_criterion: Which field name to use for sorting. E.g datetime will sort by datetime ascending
                    while -datetime will sort based on datetime descending.

            Returns:
                The part of the Klaviyo API call URL corresponding to the call parameters. E.g.
                    sort=datetime
            """
    return _SORT_ARG_TEMPLATE.substitute(sort_criterion=sort_criterion)


def get_url_since_timestamp_filter_args(since: str):
    """ Gets the filters part of the URL given the parameters

    Args:
        since: It's a Unix timestamp (UTC) that corresponds to the timestamp filter along
            with the 'greater-or-equal' option.It filters the events that happened after that
            timestamp
    """
    return get_filter_url(condition='greater-or-equal',
                          metric_id='timestamp',
                          metric_value=since)


def get_url_until_timestamp_filter_args(until: str):
    """ Gets the filters part of the URL given the parameters

    Args:
        until: It's a Unix timestamp (UTC) that corresponds to the timestamp filter along
            with the 'less-or-equal' option.It filters the events that happened before that
            timestamp
    """
    return get_filter_url(condition='less-or-equal',
                          metric_id='timestamp',
                          metric_value=until)


def get_next_page_cursor_arg(next_page: str = None):
    if next_page:
        return _PAGE_CURSOR_ARG_TEMPLATE.substitute(page_cursor=next_page)
    return ''



import hashlib


def create_md5_hash(input_str: str) -> str:
    """
    Returns: md5 hash of the input string
    """
    encoded_str = input_str.encode()
    hash_object = hashlib.md5(encoded_str)
    hash_hex = hash_object.hexdigest()
    return hash_hex


def json_object_to_list(input_json: dict, subject_key='subject', record_key='messages') -> list[dict]:
    """
    Convert Json object into a list of json records plus adding subject and unique event_id to each data record

    Parameters:
    input_json (json object): The Json object to convert with this format
        {
            "subject_key": "subject",
            "record_key": [
                json_record1,
                json_record2,
                ...
            ]
        }
    subject_key (str): The key to the subject field in input_json
    record_key (str): The key to the records list in input_json

    Returns:
    A list of json_records with subject and a generated event_id(unique to every json_record) [{subject: subject, json_data: json_record, event_id: hash_of_json_record}, ...]
    """
    json_record_list = [dict(subject=input_json[subject_key], json_data=record, event_id=create_md5_hash(str(record))) for record
                   in input_json[record_key]]
    return json_record_list



from sbcommons import utils


def test_get_value_from_path():
    input_json = {
            "a": {
                "b": {
                    "c": "value_c",
                    "d": 10
                },
                "e": "value_e"
            },
            "f": {
                "g": {
                    "h": [1, 2, 3]
                }
            }
        }
    assert utils.get_value_from_path(input_json, "a.b.c") == "value_c"
    assert utils.get_value_from_path(input_json, "a.b.d") == 10
    assert utils.get_value_from_path(input_json, "a.e") == "value_e"
    assert utils.get_value_from_path(input_json, "f.g.h") == [1, 2, 3]



import datetime as dt

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import IO
from typing import Union

import boto3
import json
import gzip


def s3_bucket(bucket: str) -> Any:
    s3_resource = boto3.resource('s3')
    return s3_resource.Bucket(name=bucket)


def get_object(bucket_name: str, key: str, stream: bool = False) -> Any:
    bucket = s3_bucket(bucket_name)
    obj = bucket.Object(key=key)
    body = obj.get().get('Body')
    return body.iter_lines() if stream else body.read()


def put_object(bucket_name: str, key: str, content: Union[bytes, IO], **kwargs):
    bucket = s3_bucket(bucket_name)
    obj = bucket.Object(key=key)
    obj.put(Body=content, **kwargs)


def write_json_list_to_s3(bucket: str, key:str, input_json: list[dict], compress=True, **kwargs):
    json_str = "\n".join(json.dumps(obj) for obj in input_json)
    content = gzip.compress(json_str.encode('utf-8')) if compress else json_str.encode('utf-8')
    put_object(bucket_name=bucket, key=key, content=content, **kwargs)


def delete_object(bucket_name: str, key: str):
    bucket = s3_bucket(bucket_name)
    obj = bucket.Object(key=key)
    obj.delete()


def delete_folder_in_bucket(bucket_name: str, prefix: str):
    bucket = s3_bucket(bucket_name)
    bucket.objects.filter(Prefix=prefix).delete()


def list_objects(bucket_name: str, path: str = None, return_object_keys: bool = True) -> List[Any]:
    bucket = s3_bucket(bucket_name)
    return [obj.key if return_object_keys else obj for obj in bucket.objects.filter(
        Prefix=path if path else ''
    )]


def list_common_object_prefixes(bucket_name: str, path: str = None,
                                delimiter: str = '/') -> List[str]:
    objects = boto3.client('s3').list_objects_v2(
        Bucket=bucket_name,
        Prefix=path,
        Delimiter=delimiter
    )
    return [cp['Prefix'] for cp in objects.get('CommonPrefixes', [])]


def generate_presigned_url(bucket_name: str, key: str, **kwargs) -> str:
    client_kwargs = {'config': kwargs.pop('Config'), 'region_name': kwargs.pop('RegionName')}
    client = boto3.client('s3', **client_kwargs)
    return client.generate_presigned_url(
        ClientMethod='get_object',
        Params={
            'Bucket': bucket_name,
            'Key': key
        },
        **kwargs
    )


def get_last_updated_object_key_in_bucket(bucket_name: str, key: str):
    """
    This function gets the last modified object by date (last modified date) in an s3 path
    """
    get_last_modified = lambda obj: int(obj.last_modified.strftime('%s'))
    objs = list_objects(bucket_name=bucket_name, path=key, return_object_keys=False)
    objs = [obj for obj in sorted(objs, key=get_last_modified)]
    last_added = objs[-1].key
    return last_added


def clear_s3_path(s3_bucket: str, s3_key: str, backup_bucket: str):
    """ Clears objects in S3 path specified by <s3_bucket> and <s3_key>. """
    s3_objects = list_objects(bucket_name=s3_bucket, path=s3_key, return_object_keys=True)
    s3_objects_dict = {key: key for key in s3_objects}

    copy_objects(source_bucket_name=s3_bucket, destination_bucket_name=backup_bucket,
                 keys=s3_objects_dict)

    for s3_key in s3_objects:
        delete_object(bucket_name=s3_bucket, key=s3_key)


def get_last_modified_timestamp(bucket_name: str, key: str) -> dt.datetime:
    """ Gets the last modified timestamp for s3://<bucket_name>/<key>.

    Wrapper function around boto3.client.head_object()

    Raises:
        ValueError if <key> evaluates to false (e.g. empty string or None).
    """
    if not key:
        raise ValueError("The <key> function parameter cannot be None or an empty string.")
    # Add forward slash after key in case it's missing
    key = key + '/' if key[-1] != '/' else key
    # Get last modified timestamp
    client = boto3.client('s3')
    last_modified_ts = client.head_object(Bucket=bucket_name, Key=key)["LastModified"]
    return last_modified_ts


def copy_objects(source_bucket_name: str, destination_bucket_name: str, keys: Dict[str, str],
                 destination_path: str = '', delete_source_objects=False,
                 filter_func: Union[Callable, None] = None, **kwargs):
    """ Copies objects from one bucket to another.

    Args:
        source_bucket_name: The name of the bucket from which we will copy objects.
        destination_bucket_name: The name of the bucket to which the objects will be copied.
        keys: A dictionary with that maps the S3 bucket key of each object we want to copy from
            <source_bucket_name> to the S3 key that we want to use when copying it under
            <destination_bucket_name>.
        destination_path: The key of the directory path in <destination_bucket_name> where we will
            copy all the objects.
        delete_source_objects: If True, each object is deleted from <source_bucket_name> once it is
            copied. This can be used to move objects.
        filter_func: A function that is used to omit certain objects in <keys> if they fulfill
            some condition. <filter_func> takes a bucket and key as an input, as well as optional
            keywords parameters (kwargs). It must return a value that can be evaluated as True or
            False. If the return value is True, the object is copied (and vice-versa).
            Defaults to None, which means all objects in <keys> are copied.
    """
    destination_bucket = s3_bucket(destination_bucket_name)

    # Ensure folder structure of destination path
    if destination_path and not destination_path.endswith('/'):
        destination_path += '/'

    for source_key, destination_key in keys.items():
        # If <filter_func> is not defined or it returns true for <source_key> then move the object
        if not filter_func or filter_func(source_bucket_name, source_key, **kwargs):
            copy_source = {
                'Bucket': source_bucket_name,
                'Key': source_key
            }

            obj = destination_bucket.Object(key=f'{destination_path}{destination_key}')
            obj.copy(copy_source)
            if delete_source_objects:
                delete_object(source_bucket_name, source_key)


def move_objects(source_bucket_name: str, destination_bucket_name: str, keys: Dict[str, str],
                 destination_path: str = '', filter_func: Union[Callable, None] = None, **kwargs):
    copy_objects(source_bucket_name, destination_bucket_name, keys, destination_path,
                 delete_source_objects=True, filter_func=filter_func, **kwargs)



from typing import OrderedDict

import configparser


def get_config_dict(config_path: str, section_key: str = None) -> OrderedDict:
    """ Reads a configuration file using configparser.

        See https://docs.python.org/3/library/configparser.html#quick-start to learn more about the
            structure of the configuration file.

        Args:
            config_path: The path to the configuration file.
            section_key: If specified the function will return only the specific section of the
                configuration specified by the <section_key>.

        Returns:
            An OrderedDict that maps the name of each section of the configuration file to an
                OrderedDict object that includes a mapping of each configuration variable name to
                its value.

                If <section_key> is specified, only one section of the configuration is returned.
     """
    cfg = configparser.ConfigParser()
    cfg.read(config_path)
    cfg_dict = cfg.__dict__['_sections']
    if section_key:
        return cfg_dict[section_key]
    return cfg_dict


def get_text_from_file(path: str) -> str:
    """ Reads an entire text file into a string variable and returns its value. """
    with open(path, 'r') as fp:
        return fp.read()



from sbcommons.crm.klaviyo.event import KlaviyoEvent
from sbcommons.crm.klaviyo.metric import KlaviyoMetric
from sbcommons.crm.klaviyo.client import KlaviyoClient
from sbcommons.crm.klaviyo.parser import KlaviyoParser



""" Class representing an event on Klaviyo """
from abc import ABC, abstractmethod
from functools import reduce
from typing import Dict
from typing import List

import pandas as pd


class KlaviyoEvent(ABC):
    """ Base Klaviyo event class"""

    @classmethod
    @property
    @abstractmethod
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        raise NotImplementedError

    @classmethod
    @property
    @abstractmethod
    def column_paths(cls):
        """ A dict that maps the path of each event json column to the name of that column. """
        raise NotImplementedError

    @staticmethod
    def get_json_col(json_dict: dict, column_path: str) -> str:
        """ Extracts the element in the json string specified by <column_path>

        Args:
            json_dict: A dict corresponding to the json file.
            column_path: The path to the element we want to extract. E.g. column_path = x.y.z if we
                want to extract value from {x: {y: {z: value}}}.

        """
        return reduce(lambda d, key: d.get(key) if d else None, column_path.split('.'), json_dict)

    @classmethod
    def events_to_df(cls, events_list: List[dict]):
        """ Converts the metric events in events_list to a pandas DataFrame.

        events_list: A list of json files where each json file represents an event.

        Returns:
            A pandas DataFrame with a number of columns extracted from each event specified by
                <column_paths>. For example, for ReceivedKlaviyoEvent we extract the following
                columns: i) campaign_name, customer_id_hash, message_sendtime.
        """
        parsed_events = []
        for event in events_list:
            parsed_event = {}
            for column_path, column_name in cls.column_paths.items():
                column_value = cls.get_json_col(event, column_path)
                parsed_event[column_name] = column_value if column_value else ''
            parsed_events.append(parsed_event)
        df_columns = list(parsed_event.keys()) if events_list else []
        df = pd.DataFrame(parsed_events, columns=df_columns)
        return df

    @classmethod
    def get_event_class_by_metric_name(cls, metric_name: str):
        """ Gets a KlaviyoEvent object (subclass) given the metric name.

        Args:
            metric_name: The name of the metric.

        Returns:
            A KlaviyoEvent subclass where the metric_name property is equal to the metric_name
            argument passed into this function. E.g. if <metric_name> == 'Received Email',
            the function returns ReceivedKlaviyoEvent.
        """
        for klaviyo_event in KlaviyoEvent.__subclasses__():
            if klaviyo_event.metric_name == metric_name:
                return klaviyo_event


class ProductTracker(KlaviyoEvent):
    """ A product tracker trigger event that sends out survey invitation.

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        original_id (str): Survey customer id hash passed from survey tracker handler run.
        datetime (str): datetime of the trigger in Klaviyo.
        profile_id (str): native customer id in Klaviyo.
    """

    column_paths = {
        'attributes.event_properties.originalId': 'original_id',
        'attributes.datetime': 'datetime',
        'relationships.profile.data.id': 'profile_id',
        'flow_id': 'flow_id'
    }

    def __init__(self, original_id: str, datetime: str, profile_id: str, flow_id: str):
        KlaviyoEvent.__init__(self)
        self.original_id = original_id
        self.datetime = datetime
        self.profile_id = profile_id
        self.flow_id = flow_id

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Product Tracker Flow'


class BrandTracker(KlaviyoEvent):
    """ A brand tracker trigger event that sends out survey invitation.

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        original_id (str): Survey customer id hash passed from survey tracker handler run.
        datetime (str): datetime of the trigger in Klaviyo.
        profile_id (str): native customer id in Klaviyo.
    """

    column_paths = {
        'attributes.event_properties.originalId': 'original_id',
        'attributes.datetime': 'datetime',
        'relationships.profile.data.id': 'profile_id',
        'flow_id': 'flow_id'
    }

    def __init__(self, original_id: str, datetime: str, profile_id: str, flow_id: str):
        KlaviyoEvent.__init__(self)
        self.original_id = original_id
        self.datetime = datetime
        self.profile_id = profile_id
        self.flow_id = flow_id


    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Brand Tracker Flow'


class ChurnTracker(KlaviyoEvent):
    """ A churn tracker trigger event that sends out survey invitation.

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        original_id (str): Survey customer id hash passed from survey tracker handler run.
        datetime (str): datetime of the trigger in Klaviyo.
        profile_id (str): native customer id in Klaviyo.
    """

    column_paths = {
        'attributes.event_properties.originalId': 'original_id',
        'attributes.datetime': 'datetime',
        'relationships.profile.data.id': 'profile_id',
        'flow_id': 'flow_id'
    }

    def __init__(self, original_id: str, datetime: str, profile_id: str, flow_id: str):
        KlaviyoEvent.__init__(self)
        self.original_id = original_id
        self.datetime = datetime
        self.profile_id = profile_id
        self.flow_id = flow_id

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Churn Tracker'


class ReceivedEmail(KlaviyoEvent):
    """ An event corresponding to the received e-mail metric

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        customer_id_hash (str): Customer id hash of the customer who got received the e-mail.
        campaign_name (str): Campaign name of the campaign the customer received by e-mail.
        message_sendtime (str): The time and date when the e-mail was received.
        profile_id (str): Klaviyo's local customer id.
        flow_id (str): The flow id of the relevant trigger flow if applicable.
        message_id (str): The message id of the e-mail.
    """

    column_paths = {
        'attributes.event_properties.Campaign Name': 'campaign_name',
        'customerhashID': 'customer_id_hash',
        'attributes.datetime': 'message_sendtime',
        'relationships.profile.data.id': 'profile_id',
        'attributes.event_properties.$flow': 'flow_id',
        'attributes.event_properties.$message': 'message_id'
    }

    def __init__(self, customer_id_hash: str, campaign_name: str, message_sendtime: str,
                 profile_id: str, flow_id: str, message_id: str):
        KlaviyoEvent.__init__(self)
        self.customer_id_hash = customer_id_hash
        self.campaign_name = campaign_name
        self.message_sendtime = message_sendtime
        self.profile_id = profile_id
        self.flow_id = flow_id
        self.message_id = message_id

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Received Email'


class OpenedEmail(KlaviyoEvent):
    """ An event corresponding to the opened e-mail metric

     Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        customer_id_hash (str): Customer id hash of the customer who got received the e-mail.
        campaign_name (str): Campaign name of the campaign the customer received by e-mail.
        message_sendtime (str): The time and date when the e-mail was received.
        profile_id (str): Klaviyo's local customer id.
        flow_id (str): The flow id of the relevant trigger flow if applicable.
        message_id (str): The message id of the e-mail.
    """

    column_paths = {
        'attributes.event_properties.Campaign Name': 'campaign_name',
        'customerhashID': 'customer_id_hash',
        'attributes.datetime': 'message_openedtime',
        'relationships.profile.data.id': 'profile_id',
        'attributes.event_properties.$flow': 'flow_id',
        'attributes.event_properties.$message': 'message_id'
    }

    def __init__(self, customer_id_hash: str, campaign_name: str, message_openedtime: str,
                 profile_id: str, flow_id: str, message_id: str):
        KlaviyoEvent.__init__(self)
        self.customer_id_hash = customer_id_hash
        self.campaign_name = campaign_name
        self.message_openedtime = message_openedtime
        self.profile_id = profile_id
        self.flow_id = flow_id
        self.message_id = message_id

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Opened Email'


class ClickedEmail(KlaviyoEvent):
    """ An event corresponding to the clicked e-mail metric.

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        metric_name (str): The name of the metric the event corresponds to.
        customer_id_hash (str): Customer id hash of the customer who got received the e-mail.
        campaign_name (str): Campaign name of the campaign the customer received by e-mail.
    """

    column_paths = {
        'attributes.event_properties.Campaign Name': 'campaign_name',
        'customerhashID': 'customer_id_hash',
    }

    def __init__(self, customer_id_hash: str, campaign_name: str):
        KlaviyoEvent.__init__(self)
        self.customer_id_hash = customer_id_hash
        self.campaign_name = campaign_name

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Clicked Email'


class Unsubscribed(KlaviyoEvent):
    """ An event corresponding to the 'Unsubscribed' metric.

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        metric_name (str): The name of the metric the event corresponds to.
    """

    column_paths = {
        'customerhashID': 'customer_id_hash',
        'attributes.datetime': 'datetime'
    }

    def __init__(self, customer_id_hash: str, event_time: str):
        KlaviyoEvent.__init__(self)
        self.customer_id_hash = customer_id_hash
        self.event_time = event_time

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Unsubscribed'


class UnsubscribedFromSMS(KlaviyoEvent):
    """ An event corresponding to the 'Unsubscribed from SMS' metric.

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        metric_name (str): The name of the metric the event corresponds to.
    """

    column_paths = {
        'customerhashID': 'customer_id_hash',
        'attributes.datetime': 'datetime'
    }

    def __init__(self, customer_id_hash: str, event_time: str):
        KlaviyoEvent.__init__(self)
        self.customer_id_hash = customer_id_hash
        self.event_time = event_time

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Unsubscribed from SMS'


class UnsubscribedFromList(KlaviyoEvent):
    """ An event corresponding to the 'Unsubscribed from List' metric.

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        metric_name (str): The name of the metric the event corresponds to.
    """

    column_paths = {
        'attributes.event_properties.List': 'list_name',
        'customerhashID': 'customer_id_hash',
        'attributes.datetime': 'datetime'
    }

    def __init__(self, customer_id_hash: str, list_name: str, datetime: str):
        KlaviyoEvent.__init__(self)
        self.customer_id_hash = customer_id_hash
        self.list_name = list_name
        self.datetime = datetime

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Unsubscribed from List'


class SubscribedToList(KlaviyoEvent):
    """ An event corresponding to the 'subscribed to list' metric.

    Attributes:
        column_paths (Dict[str, str]): A dictionary that maps different paths of the event json
            (e.g. event_properties.Campaign_Name) to the variable names in snake case
            (e.g. campaign_name).
        metric_name (str): The name of the metric the event corresponds to.
    """

    column_paths = {
        'attributes.event_properties.List': 'list_name',
        'customerhashID': 'customer_id_hash',
        'attributes.datetime': 'datetime'
    }

    def __init__(self, customer_id_hash: str, list_name: str, datetime: str):
        KlaviyoEvent.__init__(self)
        self.customer_id_hash = customer_id_hash
        self.list_name = list_name
        self.event_time = datetime

    @classmethod
    @property
    def metric_name(cls) -> str:
        """ The name of the metric the event corresponds to. """
        return 'Subscribed to List'



import datetime as dt
from logging import FileHandler
from pathlib import Path
from typing import Union


class InvalidLogNameError(Exception):
    DEFAULT_MSG = 'The log file name {LOG_FILE_NAME} should be formatted as follows: ' \
                  ' <base_file_name>__<year>_<month>_<day>.<extension>.'

    def __init__(self, msg=None, file_name=None, *args):
        self.file_name = file_name
        msg = msg if msg else self.DEFAULT_MSG.format(LOG_FILE_NAME=file_name)
        Exception.__init__(self, msg, *args)


class RotatingFileHandler(FileHandler):
    """ A FileHandler that rotates the file it writes to every day by appending a date suffix
        to the file name.

        Attributes:
            __file_path: The pathlib.Path object for the file path of the log file.
            backup_count: The maximum number of log files to keep in the same directory before
                starting to discard old log files.
    """

    def __init__(self, file_path, mode='a', encoding=None, delay=False, errors=None,
                 backup_count=0):
        """
        Open the specified file and use it as the stream for logging.
        """
        self.__file_path = Path(file_path)
        self.backup_count = backup_count
        self.rotate_file()
        FileHandler.__init__(self, file_path, mode, encoding, delay, errors)

    def rotate_file(self) -> bool:
        """ Rotates a file by adding a suffix to the file name and optionally discarding old files.

        Before we write any new logs, if the file path was last modified on a previous date,
        we rename the file by adding its last modified date as a suffix to the file name. If there
        are more than <backup_count> files saved with the same prefix, then the oldest one is
        discarded, if <backup_count> is set to be larger than 0.

        Returns:
            True if a file has been rotated, False otherwise.
        """
        last_modified_date = self.get_last_modified_date(self.__file_path)
        today_date = dt.datetime.now().date()

        # If the file doesn't exist or the last modified date is today, don't rotate
        if not last_modified_date or last_modified_date == today_date:
            return False

        # Rename old file by adding a date suffix
        self.add_date_suffix(self.__file_path, last_modified_date)

        # Discard old files with the same name if self.backup_count is set to be larger than 0
        self.discard_old_files()

        return True

    @staticmethod
    def get_last_modified_date(path: Path) -> Union[dt.date, None]:
        """ Gets last modified date from file path. Returns None if file doesn't exist. """
        if not path.exists():
            return
        return dt.datetime.fromtimestamp(path.stat().st_mtime).date()

    @staticmethod
    def add_date_suffix(path: Path, date: dt.date) -> Union[dt.date, None]:
        """ Adds the date suffix to the file name. Returns None if file doesn't exist. """
        if not path.exists():
            return
        # Creating new file name, skipping the last extension since it is automatically added when
        # calling path.rename()
        file_name, extensions = path.stem.split('.')[0], path.suffixes
        file_name += f'__{date.year}_{date.month}_{date.day}' + ''.join(extensions[:-1])
        path.rename(path.with_stem(file_name))

    @staticmethod
    def get_date_suffix(path: Path):
        """ Gets date suffix from file path.

        Args:
            path: The file path object.

        Raises:
            InvalidLogNameError if the name of the file is not formatted as follows:
                <base_name>__<year>_<month>_<day>.<extension>.
        """
        try:
            filename = path.stem.split('.')[0]
            date = filename.split('__')[1]
            year, month, day = date.split('_')
            return dt.datetime(int(year), int(month), int(day))
        except (IndexError, ValueError):
            raise InvalidLogNameError(file_name=filename)

    def discard_old_files(self):
        """ Discard old files with the same name if self.backup_count is set to be larger than 0."""
        if not self.backup_count or not self.__file_path.exists:
            return

        file_name, extensions = self.__file_path.stem.split('.')[0], self.__file_path.suffixes
        # Get log files that have the same base file name within the directory, sorted by timestamp
        log_files = sorted(self.__file_path.parent.absolute().glob(f'{file_name}__*.*'),
                           key=lambda x: self.get_date_suffix(x), reverse=True)
        # If we exceed the maximum number of files with the same name, delete the oldest files
        if len(log_files) > self.backup_count:
            discard_list = log_files[self.backup_count:]
            for path in discard_list:
                path.unlink()



from typing import Dict
from typing import List
from typing import Union

from sbcommons.crm.symplify import SymplifyClient
from sbcommons.crm.symplify import SymplifyParser
from sbcommons.crm.klaviyo import KlaviyoClient
from sbcommons.crm.klaviyo import KlaviyoParser
from sbcommons.crm.parser import TranslationEntry


SUPPORTED_CRM_PLATFORMS = ['symplify', 'klaviyo']


class InvalidCrmPlatform(Exception):
    """ Invalid/Not supported Crm Platform class"""
    def __init__(self, msg):
        Exception.__init__(msg)


class ClientFactory:
    @classmethod
    def create_client(cls, crm_platform_name: str, token: Union[str, Dict[str, str]], **kwargs) \
            -> Union[SymplifyClient, KlaviyoClient]:
        """

        Args:
            crm_platform_name: The name of the CRM platform to create a client for. E.g. Symplify,
                Klaviyo.
            token: The access token to be used to instantiate the client.
            kwargs: Dictionary of keyword arguments that are passed to the client's constructor.

        Raises:
            InvalidCrmPlatform if the crm_platform_name parameter is not one of the values in
                SUPPORTED_CRM_PLATFORMS.
        """
        if crm_platform_name == 'symplify':
            client_class = SymplifyClient
        elif crm_platform_name == 'klaviyo':
            client_class = KlaviyoClient
        else:
            raise InvalidCrmPlatform(f'CRM platform {crm_platform_name} is not supported. Select '
                                     f'one of {SUPPORTED_CRM_PLATFORMS}')
        return client_class(token=token, **kwargs)


class ParserFactory:
    @classmethod
    def create_parser(cls, crm_platform_name: str, translation_entries: List[TranslationEntry],
                      **kwargs) -> Union[SymplifyParser, KlaviyoParser]:
        """

        Args:
            crm_platform_name: The name of the CRM platform to create a client for. E.g. Symplify,
                Klaviyo.
            translation_entries: A list of translation entry objects used to parse the results
                    of the SQL query to the format required by the CRM tool.

        Raises:
            InvalidCrmPlatform if the crm_platform_name parameter is not one of the values in
                SUPPORTED_CRM_PLATFORMS.
        """
        if crm_platform_name == 'symplify':
            parser_class = SymplifyParser
        elif crm_platform_name == 'klaviyo':
            parser_class = KlaviyoParser
        else:
            raise InvalidCrmPlatform(f'CRM platform {crm_platform_name} is not supported. Select '
                                     f'one of {SUPPORTED_CRM_PLATFORMS}')
        return parser_class(translation_entries=translation_entries, **kwargs)



import unittest
import sbcommons.extract_utils.utils as utils


def test_create_md5_hash():
    # Test case 1: Known string and its SHA-256 hash
    input_str = "hello"
    expected_hash = "5d41402abc4b2a76b9719d911017c592"
    assert utils.create_md5_hash(input_str) == expected_hash, "Create md5 for 'hello' failed"

    # Test case 2: Empty string
    input_str = ""
    expected_hash = "d41d8cd98f00b204e9800998ecf8427e"
    assert utils.create_md5_hash(input_str) == expected_hash, "Create md5 for blank string failed"


def test_create_md5_hash_uniqueness():
    input_str = "A B"*1000
    assert utils.create_md5_hash(input_str) == utils.create_md5_hash(input_str), "Create_hash function uniqueness failed"


def test_json_object_to_list():
    input_json = {
        "subject": "Test Subject",
        "messages": [
            {"message": "Message 1"},
            {"message": "Message 2"}
        ]
    }
    expected_output = [
        {
            "subject": "Test Subject",
            "json_data": {"message": "Message 1"},
            "event_id": utils.create_md5_hash(str({"message": "Message 1"}))
        },
        {
            "subject": "Test Subject",
            "json_data": {"message": "Message 2"},
            "event_id": utils.create_md5_hash(str({"message": "Message 2"}))
        }
    ]
    result = utils.json_object_to_list(input_json)
    assert result == expected_output, "Converting json object to json list with default subject and record failed"


def test_empty_json_object_to_list():
    input_json = {
        "subject": "Test Subject",
        "messages": []
    }
    expected_output = []
    result = utils.json_object_to_list(input_json)
    assert result == expected_output, "Converting empty jason object to list failed"


def test_json_object_to_list_with_keys():
    # Try json_object_to_list with different keys
    input_json = {
        "topic": "Different Subject",
        "entries": [
            {"data": "Entry 1"},
            {"data": "Entry 2"}
        ]
    }
    expected_output = [
        {
            "subject": "Different Subject",
            "json_data": {"data": "Entry 1"},
            "event_id": utils.create_md5_hash(str({"data": "Entry 1"}))
        },
        {
            "subject": "Different Subject",
            "json_data": {"data": "Entry 2"},
            "event_id": utils.create_md5_hash(str({"data": "Entry 2"}))
        }
    ]
    result = utils.json_object_to_list(input_json, subject_key='topic', record_key='entries')
    assert result == expected_output, "Converting json object to list with different keys failed"



import json
from typing import Dict

import boto3
from botocore.config import Config


def invoke_with_extended_timeout(function_name: str, event: Dict, invocation_type: str = 'Event'):
    config = Config(read_timeout=900, retries={'max_attempts': 1})
    return invoke(function_name, event, invocation_type, config)


def invoke(function_name: str, event: Dict, invocation_type: str = 'Event', config: Config = None):
    lambda_resource = boto3.client('lambda', config=config)
    return lambda_resource.invoke(
        FunctionName=function_name,
        InvocationType=invocation_type,
        Payload=json.dumps(event)
    )



import logging
import requests
from requests.structures import CaseInsensitiveDict
from time import sleep


from sbcommons.crm.client import CrmClient


class SymplifyClient(CrmClient):
    """ A client class to interact with Symplify's API.

    Attributes:
        token (str): Token needed to access Symplify's API. Inherited from CrmClient.
        customer_id (str): Customer here refers to our company. This id corresponds to our company
            account on Symplify.
        list_id (int): Identifier of the list to use for pushing/retrieving customers.
        delimiter (str): Delimiter to use to separate column fields in exports.
        session (requests.Session): The session object for the Symplify connection. Inherited from
            CrmClient.
        rate_limit (int): Amounts of second to wait after getting a rate limited error (HTTP 429),
            in order to make a successful request again. Inherited from CrmClient.
        max_rate_limit_retries (int): Number of times to retry making a request if a rate limited
            error occurs (status code 429/Too Many Requests). Inherited from CrmClient.
    """

    # Symplify's REST API base URL
    BASE_URL = 'http://www.carmamail.com:80/rest/'

    def __init__(self, token: str, customer_id: str, list_id: int = None, delimiter: str = '|',
                 rate_limit=60, max_rate_limit_retries=2, logger: logging.Logger = None):
        CrmClient.__init__(self, token=token, rate_limit=rate_limit,
                           max_rate_limit_retries=max_rate_limit_retries)
        self.customer_id = customer_id
        self.list_id = list_id
        self.delimiter = delimiter
        # If logger argument is None, use default logger
        self._logger = logger or logging.getLogger(__name__)

    @property
    def logger(self) -> logging.Logger:
        """ Returns the logger object to be used by the instance. """
        return self._logger

    def connect(self):
        """ Performs the necessary actions required to be able to make requests to Symplify.

        This method is automatically called when this class is used in a context manager i.e. when
        __enter__ is called.
        """
        self.session.mount('https://', self._retry_adapter(retries=3, backoff_factor=4))
        self.session.headers = self._build_base_header()

    def close(self):
        """ Performs necessart actions to terminate the connection with Symplify. """
        self.session.close()

    def _build_base_header(self) -> CaseInsensitiveDict:
        """ Builds a dictionary that includes the header's basic elements for making requests. """
        return CaseInsensitiveDict({
            'Accept': 'application/json',
            'X-Carma-Authentication-Token': self.token
        })

    def _build_url(self, endpoint: str) -> str:
        """ Builds the URL that will be used for requests to the Symplify API.

        Args:
            endpoint: The part of the request URL following <server>/rest/<customerId>.

        Returns:
            A string representing the URL.
        """
        return f'{self.BASE_URL}{self.customer_id}/{endpoint}'

    def create_import(self, import_type: str = 'ADD', identity_column: str = 'originalId') -> str:
        """ Calls the Symplify API to create an id for importing customers into the list.

        Args:
            import_type: This can be either 'ADD' or 'REPLACE'. Using 'ADD' will add new customers
                to the list, updating the attributes of those that already exist. 'REPLACE' will
                delete all customers from the list and replace them with the new ones.
            identity_column: The name of the column in the imported data that is used to distinguish
                between customers. If we import a customer with an originalId that already exists in
                the list and we use the 'ADD' import type, then we will simply update the attributes
                of that customer.

        Returns:
            The identifier to use for importing into the list.
        """
        url = self._build_url(f'lists/{self.list_id}/imports')
        payload = {
            'delimiter': f'{ord(self.delimiter)}',
            'encoding': 'UTF8',
            'type': import_type,
            'identityColumn': identity_column
        }
        return self._request(method='POST', url=url, json=payload, verbose=True).json()['id']

    def post_list_async(self, import_type: str, list_data: bytes) -> int:
        """ Posts a list of customers to Symplify - asynchronous call.

            import_type: This can be either 'ADD' or 'REPLACE'. Using 'ADD' will add new customers
                to the list, updating the attributes of those that already exist. 'REPLACE' will
                delete all customers from the list and replace them with the new ones.
            identity_column: The name of the column in the imported data that is used to distinguish
                between customers. If we import a customer with an originalId that already exists in
                the list and we use the 'ADD' import type, then we will simply update the attributes
                of that customer.

            Returns:
                The batch id for the batch import job. The batch id can be used to track the
                    asynchronous call and get its status using the <check_batch_status> method.
        """
        import_id = self.create_import(import_type=import_type)
        url = self._build_url(endpoint=f'lists/{self.list_id}/recipients/{import_id}')
        self.session.headers['Content-Type'] = 'text/csv'
        return self._request(method='POST', url=url, data=list_data, verbose=True).json()['batchId']

    def post_list_sync(self, list_data: bytes, import_type: str = 'ADD') -> bool:
        """ Posts a list of customers to Symplify - synchronous call.

            import_type: This can be either 'ADD' or 'REPLACE'. Using 'ADD' will add new customers
                to the list, updating the attributes of those that already exist. 'REPLACE' will
                delete all customers from the list and replace them with the new ones.
            identity_column: The name of the column in the imported data that is used to distinguish
                between customers. If we import a customer with an originalId that already exists in
                the list and we use the 'ADD' import type, then we will simply update the attributes
                of that customer.

            Returns:
                True if the job succeeded. False otherwise.
        """
        batch_id = self.post_list_async(import_type=import_type, list_data=list_data)
        symplify_batch_status = self.check_batch_status(batch_id)
        while symplify_batch_status != 'FINISHED' and symplify_batch_status != 'ERROR':
            sleep(self.rate_limit)
            symplify_batch_status = self.check_batch_status(batch_id)
        job_succeeded = True if symplify_batch_status == 'FINISHED' else False
        return job_succeeded

    def check_batch_status(self, batch_id: int):
        url = self._build_url(endpoint=f'batches/{batch_id}')
        self.session.headers['Content-Type'] = 'application/json'
        return self._request(method='GET', url=url, verbose=True).json()['status']

    def clean_list(self) -> bool:
        """ Clears all customer data from a list.

        Returns:
            True if the list was successfully cleaned. False otherwise.
        """
        payload_type = 'REPLACE'
        header = 'originalId'
        str_data = f'{header}'

        self.logger.info(f'Posting 0 rows to list {self.list_id}.')
        bytes_data = str_data.encode('utf-8')
        job_succeeded = self.post_list_sync(import_type=payload_type, list_data=bytes_data)
        return job_succeeded

    def post_survey_candidates(self, **kwargs):
        """ Wrapper around post_list_sync. Used to achieve polymorphism with other CrmClient
        subclasses such as KlaviyoClient. We might use different methods to post survey candidates
        in different CRM systems. """
        return self.post_list_sync(**kwargs)



from typing import Any

import boto3


def dms_client(region: str) -> Any:
    return boto3.client('dms', region_name=region)


def describe_replication_task(region: str, is_single: bool = True, **kwargs):
    client = dms_client(region)
    replication_tasks = client.describe_replication_tasks(**kwargs)
    return replication_tasks['ReplicationTasks'][0] if is_single else replication_tasks


def start_replication_task(region: str, task_arn: str, task_type: str, **kwargs):
    client = dms_client(region)
    client.start_replication_task(
        ReplicationTaskArn=task_arn,
        StartReplicationTaskType=task_type,
        **kwargs
    )



import ast
import logging
from functools import reduce
import re
from typing import List, Any


def evaluate_recursively(obj, logger: logging.Logger = None):
    """ Evaluates a python object recursively.

        Example:
            "{'give_away': {'amount': 500}}"
            returns
            {give_away': {'amount': 500}}

        Args:
            obj: Any python object
            logger: Object for logging.

        Return:
            The actual evaluated python object down to the root element.
    """
    try:
        if isinstance(obj, list):
            data = [evaluate_recursively(el, logger) for el in obj]
        elif isinstance(obj, dict):
            data = obj
            for k, v in data.items():
                data[k] = evaluate_recursively(v, logger)
        elif isinstance(obj, tuple):
            data = (evaluate_recursively(el, logger) for el in obj)
        elif isinstance(obj, (int, bool, float)):
            return obj
        else:
            data = ast.literal_eval(obj)
            if isinstance(data, (list, dict, tuple)):
                data = evaluate_recursively(data)
        return data
    except (ValueError, SyntaxError) as e:
        # If we got a Value error try adding quotes to the object in case it is a string
        if isinstance(e, ValueError):
            try:
                data = ast.literal_eval("'" + obj + "'")
                return data
            except SyntaxError:
                pass
        # If adding quotes did not help then log a warning but return object as is
        if logger:
            logger.warning(f"Warning: object could not be parsed: {repr(e)}")
        return obj


def get_field_from_included(data_list, included_list, field_path):
    field_name = field_path.split('.')[-1]
    for data_dict_idx, data_dict in enumerate(data_list):
        for included_dict in included_list:
            # fetch the value if the kaviyo ids match
            if data_dict['relationships']['profile']['data']['id'] == included_dict['id']:
                result = rget(included_dict, field_path)
                if result:
                    data_list[data_dict_idx][field_name] = result
                break
    return data_list


def get_value_from_path(data: dict, path: str):
    """ Helper function to get nested field from dictionary.
    Args:
        data: The dictionary with the data at hand.
        path: The path inside the dictionary that we want to access. Each field is sperated by ".".
        """
    if not isinstance(data, dict):
        raise TypeError(f"parameter data is of type {type(data)} instead of dict")
    keys = path.split(".")
    result = data
    try:
        for key in keys:
            result = result[key]
        return result
    except KeyError:
        return None


def rget(obj, path):
    return reduce(lambda d, key: d.get(key) if d else None, path.split('.'), obj)


def remove_email_info_from_text(text: str):
    """ Given some text we remove any email information and replace it with *
    Parameters:
        text: The text for which we want to remove email information """

    email_pattern = r'[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}'
    return re.sub(pattern=email_pattern,
                  repl='*****',
                  string=text)
def chunk_list(x: List[Any], n: int):
    """Yield successive n-sized chunks from x."""
    for i in range(0, len(x), n):
        yield x[i:i + n]



from sbcommons.messaging import webhooks


def send_message(webhook_url: str, title: str, msg: str) -> bool:
    data = {
        "Text": msg,
        "TextFormat": "markdown",
        "Title": title
    }

    webhooks.post_to_webhook(service='teams', webhook_url=webhook_url, json=data)



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="sbcommons",
    version="1.1.15",
    author="Haypp Group",
    author_email="data@hayppgroup.com",
    description="Packages shared between several Data related systems in Haypp Group",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Snusbolaget",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "boto3>=1.24.35",
        "botocore>=1.27.35 ",
        "urllib3>=1.26.10",
        "requests>=2.28.1",
        "aws-secretsmanager-caching>=1.1.1.5",
    ]
)



import copy
import json
import logging
import time
from typing import Dict, Callable, Any
from typing import List
import requests
import urllib.parse

from requests.structures import CaseInsensitiveDict

from sbcommons.crm.klaviyo.url import (get_next_page_cursor_arg,
                                       get_filter_url,
                                       get_url_since_timestamp_filter_args,
                                       get_url_until_timestamp_filter_args,
                                       get_field_arg)
from sbcommons.utils import get_field_from_included, get_value_from_path, chunk_list
from sbcommons.crm.client import CrmClient
from sbcommons.crm.klaviyo.metric import KlaviyoMetric
from sbcommons.crm.klaviyo.metric import KlaviyoMetricError
from string import Template


class KlaviyoClientError(Exception):
    """ Klaviyo client exception class"""

    def __init__(self, msg):
        super(KlaviyoClientError, self).__init__(msg)


class KlaviyoClient(CrmClient):
    """ A client class to interact with Klaviyo's API.

    Attributes:
        token (Dict[str, str]): A Dict with two keys, the PUBLIC_API_KEY and PRIVATE_API_KEY
            required to make requests to Klaviyo. Inherited from CrmClient.
        list_id (int): Identifier of the list to use for pushing/retrieving customers.
        session (requests.Session): The session object for the Klaviyo connection. Inherited from
            CrmClient.
        rate_limit (int): Amounts of second to wait after getting a rate limited error (HTTP 429),
            in order to make a successful request again. Inherited from CrmClient.
        max_rate_limit_retries (int): Number of times to retry making a request if a rate limited
            error occurs (status code 429/Too Many Requests). Inherited from CrmClient.
    """

    _KLAVIYO_IDENTIFY_URL = 'https://a.klaviyo.com/client/profiles/?company_id='
    _KLAVIYO_UNSUBSCRIBE_URL = 'https://a.klaviyo.com/api/v1/people/exclusions'
    _KLAVIYO_TRACK_URL = 'https://a.klaviyo.com/client/events/?company_id='

    _KLAVIYO_GET_GROUP_MEMBERS_URL_TEMPLATE = ('https://a.klaviyo.com/api/lists/'
                                               '{list_id}/profiles/?fields[profile]=email&'
                                               'page[size]=100')
    _KLAVIYO_GET_SEGMENT_MEMBERS_URL_TEMPLATE = ('https://a.klaviyo.com/api/segments/{segment_id}'
                                                 '/profiles/?page[size]=100')
    _KLAVIYO_LIST_MEMBERS_URL = Template('https://a.klaviyo.com/api/lists/$id/relationships/'
                                         'profiles/?fields[profile]=email&page[size]=100')
    _KLAVIYO_SUBSCRIBE_PROFILES = 'https://a.klaviyo.com/api/profile-subscription-bulk-create-jobs/'
    _KLAVIYO_UPDATE_PROFILE_URL = 'https://a.klaviyo.com/api/profile-import/'
    _KLAVIYO_REQUEST_PROFILE_DELETION_URL = 'https://a.klaviyo.com/api/data-privacy-deletion-jobs/'
    _KLAVIYO_FLOWS_URL = "https://a.klaviyo.com/api/flows/?page[size]=50"
    _KLAVIYO_BULK_UPSERT_PROFILES_URL = 'https://a.klaviyo.com/api/profile-bulk-import-jobs/'
    _KLAVIYO_LIST_MEMBERS_COUNT = ('https://a.klaviyo.com/api/lists/id?additional-fields[list]'
                                   '=profile_count')
    _KLAVIYO_SEGMENT_MEMBERS_COUNT = ('https://a.klaviyo.com/api/segments/'
                                      'id?additional-fields[segment]=profile_count')
    EXCLUSION_REASONS = [
        'unsubscribed', 'bounced', 'invalid_email', 'reported_spam', 'manually_excluded'
    ]

    def __init__(self, token: Dict[str, str], list_id: str = None, segment_id=None, rate_limit=60,
                 max_rate_limit_retries=2, logger: logging.Logger = None):
        CrmClient.__init__(self, token=token, rate_limit=rate_limit,
                           max_rate_limit_retries=max_rate_limit_retries)
        self.list_id = list_id
        self.segment_id = segment_id
        # If logger argument is None, use default logger
        self._logger = logger or logging.getLogger(__name__)

    @property
    def klaviyo_identify_url(self):
        return self._KLAVIYO_IDENTIFY_URL

    @property
    def klaviyo_unsubscribe_url(self):
        return self._KLAVIYO_UNSUBSCRIBE_URL

    @property
    def klaviyo_track_url(self):
        return self._KLAVIYO_TRACK_URL

    @property
    def logger(self) -> logging.Logger:
        """ Returns the logger object to be used by the instance. """
        return self._logger

    def connect(self):
        """ Performs the necessary actions required to be able to make requests to Klaviyo.

        This method is automatically called when this class is used in a context manager i.e. when
        __enter__ is called.
        """
        self.session.mount('https://', self._retry_adapter(retries=3, backoff_factor=4))
        self.session.headers = self._build_base_header()

    def close(self):
        """  Performs necessary actions to terminate the connection with Klaviyo. """
        self.session.close()

    def _build_base_header(self) -> CaseInsensitiveDict:
        return CaseInsensitiveDict({
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': f'Klaviyo-API-Key {self.get_private_api_key()}',
            'revision': '2024-07-15'
        })

    def get_public_api_key(self):
        return self.token['PUBLIC_API_KEY']

    def get_private_api_key(self):
        return self.token['PRIVATE_API_KEY']

    def get_public_private_api_key(self):
        return self.get_public_api_key(), self.get_private_api_key()

    def get_all_metrics(self) -> List[Dict]:
        """ Returns a list of all Klaviyo metrics as dictionaries.

            Returns:
                A list of dictionaries, where each dictionary represents a metric.
        """
        data = []
        page_cursor = None
        while not data or page_cursor:
            next_page_arg = get_next_page_cursor_arg(next_page=page_cursor)
            url = f'{KlaviyoMetric.metrics_info_url}?{next_page_arg}'
            res = self._request(method='GET', url=url).json()
            page_cursor = res['links']['next']
            data.extend(res['data'])
        return data

    def get_events(self, since_ts: str,
                   end_ts: str,
                   metric_name: str,
                   include: str = 'profile',
                   obj_to_get_from_inc: List = None):
        """ Gets all Klaviyo events for <metric_name>, during the period (<since_ts>, <end_ts>).

            Args:
                since_ts: A unix timestamp with the seconds since epoch in UTC. Marks the beginning
                    of the window for which we will extract event data.
                end_ts: A unix timestamp with the seconds since epoch in UTC. Marks the end of the
                    window for which we will extract event data. If set to None, then we extract
                    all data available since <since_ts>.
                metric_name: Name of the Klaviyo metric.
                include: if any additional fields are needed (e.g profile)
                obj_to_get_from_inc: A list of objects to get from the included field

            Returns:
                A list of event dictionaries.

            Raises:
                KlaviyoError if the metric with the specified name is not found.
        """
        # Get id of metric given the name
        metric_id = self.get_metric_id(metric_name=metric_name)

        # Configure URL and URL arguments to get events for <metric_id> since <since_ts>
        url_base = KlaviyoMetric.get_metric_events_url(metric_id)
        url_args = ''
        if since_ts:
            url_args += f"%2C{get_url_since_timestamp_filter_args(since_ts)}"
        if end_ts:
            url_args += f"%2C{get_url_until_timestamp_filter_args(end_ts)}"

        url = url_base + url_args
        url += f'&{KlaviyoMetric.include_template.substitute(value=include)}'
        # Get all events in batches
        event_list = []
        next_page = None
        while not event_list or next_page:
            # Get data with Klaviyo client
            res = self._request(method='GET', url=url).json()
            # Add data to list
            if obj_to_get_from_inc:
                for ob in obj_to_get_from_inc:
                    event_list.extend(get_field_from_included(res['data'],
                                                              res['included'],
                                                              f'attributes.properties.{ob}'))
            else:
                event_list.extend(res['data'])

            # If there are no more pages then break the loop
            if res['links']['next'] is None:
                break
            next_page = res['links']['next']
            url = next_page
            # Sleep for a bit - Klaviyo allows 350 calls per second before a rate limited error
            time.sleep(0.01)

        return event_list

    def get_metric_id(self, metric_name: str) -> str:
        """ Calls the Klaviyo API to get the metric id given a metric name.

        k_client: KlaviyoClient object for making the call.

        Returns:
            The metric id character string.

        Raises:
            KlaviyoError if the metric with the specified name is not found.
        """
        # Iterate over data and return correct metric_id
        metric_list = self.get_all_metrics()
        for metric in metric_list:
            if metric['attributes']['name'] == metric_name:
                return metric['id']
        raise KlaviyoMetricError(f'Metric with name {metric_name} not found.')

    def get_flow_id(self, flow_name: str) -> str:
        """ Calls the Klaviyo API to get the flow id given a flow name.

        k_client: KlaviyoClient object for making the call.

        Returns:
            The flow id character string.

        Raises:
            KlaviyoError if the flow with the specified name is not found.
        """
        encoded_flow_name = urllib.parse.quote(flow_name)
        flow_filter = f"&filter=equals(name,'{encoded_flow_name}')"
        url = f'{self._KLAVIYO_FLOWS_URL}{flow_filter}'
        res = self._request(method='GET', url=url).json()
        data = res.get('data', [])
        if data:
            # Extract the 'id' from 'data'
            flow_id = data[0].get('id')
            return flow_id
        else:
            raise KlaviyoMetricError(f'Flow with name {flow_name} not found.')

    def get_global_exclusions(self, count: int = 5000, max_pages: int = 50,
                              reason: str = '') -> List[Dict]:
        """ Returns global exclusions/unsubscribes.

         Global exclusions are distinct from list exclusions in that these email addresses will not
         receive any emails from any list.

         https://developers.klaviyo.com/en/reference/get-global-exclusions

        Args:
            count: The number of results to return.
            max_pages: The maximum amount of pages that is expected to contain all the unsubscribes,
                with <count> unsubscribes in each page.
            reason: The exclusion reason based on which we can filter which types of exclusions we
                want to retrieve by calling get_global_exclusions. The value of this should be
                a member of the EXCLUSION_REASONS list attribute of the KlaviyoClient class.

        Returns:
            A list of dictionaries where each dictionary represents a global unsubscribe.

        Raises:
            KlaviyoClientError if the exclusion reason is a non empty string that is not in the
                EXCLUSION_REASONS list attribute.
        """
        if reason and reason not in self.EXCLUSION_REASONS:
            raise KlaviyoClientError('The value provided to the get_global_exclusions call for'
                                     'the argument <reason> is not a member of the '
                                     'EXCLUSION_REASONS list attribute in KlaviyoClient.')
        unsubscribes_list = []
        for page_i in range(max_pages):
            url_args = (f'reason={reason}&count={count}&page={page_i}'
                        f'&api_key={self.get_private_api_key()}')
            url = f'{self.klaviyo_unsubscribe_url}?{url_args}'
            response = self._request(method="GET", url=url)
            unsubscribes_list.extend(response.json()['data'])
        return unsubscribes_list

    def update_profiles(self, update_data: List[Dict], company_id: str, store: str = 'N/A'):
        """ Updates the profile attributes specified in <update_data>

        Args:
            update_data: A list of dictionaries where each dictionary represents a profile and
                includes the attributes that are going to be updated and their corresponding values.
                In the dictionary we additionally expect an extra field called 'email' we use that
                for logging purposes in case the update fails. We then delete that from the dictio
                nary
            company_id: The public API key/ site ID to be used.
            store: Store name for which we run the update. This is used for logging failed updates.
        """

        def format_payload_json(record, **kwargs):
            email = record.pop('$email')
            record['email'] = email
            f = {
                'data':
                    {
                        'type': 'profile',
                        'attributes': {
                            "email": email,
                            'properties': record}
                    }
            }

            data = json.dumps(f, indent=4)
            return email, data

        self._klaviyo_request(update_data, store, url=f'{self._KLAVIYO_IDENTIFY_URL}{company_id}',
                              method=format_payload_json)

    def update_location(self, update_data: List[Dict], store: str = 'N/A'):
        """ Updates location information for a given user

        Args:
            update_data: A list of dictionaries where each dictionary contains information regarding
            the customer location (zip code, city e.t.c).

            store: Store name for which we run the update. This is used for logging failed updates.
        """

        def format_payload_json(record, **kwargs):
            email = record.pop('$email')
            f = {
                "type": "profile",
                "attributes": {
                    "email": email,
                    "location": {
                        k: v for k, v in record.items()
                    }
                }
            }
            return f

        successful = self._start_bulk_import_profile_job(
            format_method=format_payload_json,
            profiles=update_data
        )

        return successful

    def create_event_in_metric(self, update_data: List[Dict], store: str = 'N/A'):
        """ Updates a Klaviyo metric with event attributes specified in <update_data>

        Args:
            update_data: A list of dictionaries where each dictionary represents a customer's email,
            event name, metric name, token and also includes the attributes that are going to be
            updated and their corresponding values.
            store: Store name for which we run the update. This is used for logging failed updates.
        """

        def format_payload_json(record, **kwargs):
            email = record.pop('email')
            event_name = record.pop('event_name')
            f = {
                'data': {
                    'type': 'event',
                    'attributes': {
                        'properties': record,
                        'metric': {
                            'data': {
                                "type": "metric",
                                'attributes':
                                    {'name': event_name,
                                     }
                            }
                        },
                        'profile': {'data':
                                        {'type': 'profile',
                                         'attributes': {
                                             'email': email,
                                         }
                                         }
                                    }
                    }
                }
            }

            data = json.dumps(f, indent=4)
            return email, data

        self._klaviyo_request(update_data, store,
                              url=self._KLAVIYO_TRACK_URL + self.token['PUBLIC_API_KEY'],
                              method=format_payload_json)

    def _klaviyo_request(self, update_data: List[Dict], store: str,
                         url: str, method: Callable, **kwargs):
        """ Calls a request post method

               Args:
                   update_data: A list of dictionaries where each dictionary represents a
                       customer's profile and includes the attributes that are going to
                       be updated and their corresponding values.
                   store: Store name for which we run the update. This is used for logging
                       failed updates.
                   url: endpoint /api url
                   method: a callable function that returns a json formatted str.
               """
        for record in update_data:
            email, data = method(record, **kwargs)
            try:
                self._request(method='POST', url=url, data=data)
            except Exception as e:
                self.logger.exception(f"Failed to update {email} for {store} got exception {e}")

    def get_segment_members(self, extra_fields: List[tuple] = None) -> List[Dict]:
        """Get all members from the segment
        Args:
            'extra_fields': A list of the extra fields to return if any. The list should consist of
            tuples where each tuple is (field_name, field_path). The field name is the name which
            will be returned by the function and the field path should be the path that needs to be
            followed in order to fetch the value in the JSON seperated by '.'. So if for example we
            want to fetch a field named 'user_id' which lies under attributes.properties.user id
            then extra_fields = ('user_id','attributes.properties.user_id').
        Returns:
            A list of dictionaries where each dictionary corresponds to a member of the segment. E.g.
            [{'id': '01GCY4YJKP67R3ZEHAVEYC1KXK', 'email': 'test1@hayppgroup.com'},
             {'id': '01GD38S642GQW6TR46S30KBYNQ', 'email': 'test2@hayppgroup.com'}]
        """
        get_members_url = self._KLAVIYO_GET_SEGMENT_MEMBERS_URL_TEMPLATE.format(
            segment_id=self.segment_id
        )
        url = get_members_url
        # Use pagination to iterate over all members
        segment_members = []

        next_page = None
        while not segment_members or next_page:
            json_response = self._request(method='GET', url=url).json()
            if len(json_response['data']) == 0:
                break
            for data in json_response['data']:
                if not extra_fields:
                    segment_members.append({'id': data['id'],
                                            'email': data['attributes']['email']})
                else:
                    extra_fields_dict = {field[0]: get_value_from_path(data, field[1])
                                         for field in extra_fields}

                    segment_members.append({'id': data['id'],
                                            'email': data['attributes']['email'],
                                            **extra_fields_dict})
            next_page = json_response['links']['next']
            url = next_page

        return segment_members

    def get_list_members(self) -> List[Dict]:
        """ Gets all members from the list.

        Returns:
            A list of dictionaries where each dictionary corresponds to a member of the list. E.g.
            [{'id': '01GCY4YJKP67R3ZEHAVEYC1KXK', 'email': 'test1@hayppgroup.com'},
             {'id': '01GD38S642GQW6TR46S30KBYNQ', 'email': 'test2@hayppgroup.com'}]
        """
        get_members_url = self._KLAVIYO_GET_GROUP_MEMBERS_URL_TEMPLATE.format(
            list_id=self.list_id
        )
        url = get_members_url
        # Use pagination to iterate over all members
        list_members = []

        next_page = None
        while not list_members or next_page:
            json_response = self._request(method='GET', url=url).json()
            if len(json_response['data']) == 0:
                break
            list_members.extend([{'id': data['id'], 'email': data['attributes']['email']}
                                 for data in json_response['data']])
            next_page = json_response['links']['next']
            url = next_page
        return list_members

    def remove_list_members(self, list_id, members_payload):
        """ Removes specified members from a list.

        Args:
             list_id: the id of the list from which to remove members.
             members_payload: A dictionary where the key is a type of customer identifier (e.g.
             "emails"), and the value is a list with the identifier values (e.g. e-mail addresses)
             of the customers we want to delete. E.g.

             {"emails": ['test@hayppgroup.com', 'test2@hayppgroup.com']}
        """

        def _preprocess_payload(payload):
            return {'data': [{'type': 'profile',
                              **member} for member in payload]}

        url = self._KLAVIYO_LIST_MEMBERS_URL.substitute(id=list_id)
        return self._request(method='DELETE', url=url, json=_preprocess_payload(members_payload))

    def clean_list(self, list_id) -> bool:
        """ Clears the list.

            Returns:
                True if no exception is raised while clearing the list.
        """
        records = self.get_list_members()
        if not records:
            # Return because if we make a request with no records it will cause a 400 Bad Request
            return True
        self.remove_list_members(list_id=list_id,
                                 members_payload=records)
        return True

    def post_list_sync(self, list_data: List, import_type: str = 'ADD'):
        """ Posts a list of customers to the Klaviyo list.

        Args:
            import_type: This can be either 'ADD' or 'REPLACE'. Using 'ADD' will add new customers
                to the list, while 'REPLACE' will first clear the list and then add those new
                customers.
            list_data: A dictionary with two keys list_id and a  key ("profiles") which is
            mapped to a list of dictionaries where each dictionary corresponds to the customer
            profile we want to add to the list. The profile must have an identifier such as a mobile
             phone number or e-mail address to be added to the list. Uses the Klaviyo API function
             described in https://developers.klaviyo.com/en/reference/add-members.

        Returns:
            We return True if the call is successful. False, otherwise.

        Raises:
            AssertionError if import_type is not 'REPLACE' or 'ADD'.
        """

        def format_payload_json(records, **kwargs):
            f = {
                "data": {
                    "type": "profile-subscription-bulk-create-job",
                    "attributes": {
                        "profiles": {
                            "data": [
                                {
                                    "attributes": {
                                        **record
                                    }
                                }
                            ] for record in records
                        }
                    },
                    "relationships": {
                        "list": {
                            "data": {
                                "type": "list",
                                "id": list_id
                            }
                        }
                    }
                }
            }

            data = json.dumps(f, indent=4)
            return data

        list_id = self.list_id
        assert import_type in ('REPLACE', 'ADD')
        if import_type == 'REPLACE':
            self.clean_list(list_id=list_id)
        if not len(list_data):
            return True
        url = self._KLAVIYO_SUBSCRIBE_PROFILES
        self._request(method='POST', url=url, json=format_payload_json(list_data))
        return True

    def post_survey_candidates(self, **kwargs):
        """ Wrapper around create_event_in_metric. Used to achieve polymorphism with other CrmClient
        subclasses such as SymplifyClient. We do this because we might use different methods to post
        survey candidates in different CRM systems. """
        self.create_event_in_metric(**kwargs)
        return True

    def request_customer_deletion(self, email):
        """ Requests Klaviyo to delete  customers given their e-mails.
        Args:
            email: The email of the customer we want to delete from Klaviyo."""

        payload = {
            "data": {
                "type": "data-privacy-deletion-job",
                "attributes":
                    {"profile":
                        {"data": {
                            "type": "profile",
                            "attributes": {"email": email}
                        }}}
            }}

        url = self._KLAVIYO_REQUEST_PROFILE_DELETION_URL
        self._request(method='POST', url=url, json=payload)
        return True

    def bulk_subscribe_profiles(self, profiles: List[Any], list_id: str = None) -> bool:
        """
        Subscribe profiles to email communication from klaviyo in bulk.

        Rate limits:
            Burst: 75/s
            Steady: 700/m

        See klaviyo api docs:
            https://developers.klaviyo.com/en/reference/bulk_subscribe_profiles
        Args:
            profiles: A list of dictionary entries with user profile information from klaviyo.
                        Each entry must contain 'id' and 'email' as keys.
            list_id: The id of a list in klaviyo which all profiles will be added to.
                        This list_id is optional.

        Returns:
            True if method executes without failure
        """
        job_type = 'profile-subscription-bulk-create-job'
        url = self._KLAVIYO_SUBSCRIBE_PROFILES

        def format_profile(profile):
            email = profile['email']
            profile_id = profile['id']
            data = {
                "type": "profile",
                "id": profile_id,
                "attributes": {
                    "email": email,
                    "subscriptions": {
                        "email": {
                            "marketing": {
                                "consent": "SUBSCRIBED"
                            }
                        }
                    }
                }
            }
            return data

        chunks = chunk_list(profiles, 1000)

        for chunk in chunks:
            payload = self._get_payload_for_bulk_job(
                job_type=job_type,
                format_method=format_profile,
                profiles=chunk,
                list_id=list_id
            )

            response = self._request(method='POST', url=url, json=payload)
            response.raise_for_status()

        return True

    def bulk_import_profiles(self, profiles: List[Any], list_id: str = None) -> bool:
        """
        Bulk upsert profiles to klaviyo.

        Rate limits:
            Burst: 10/s
            Steady: 150/m

        See klaviyo api docs:
            https://developers.klaviyo.com/en/reference/spawn_bulk_profile_import_job

        Args:
            profiles: A list of dictionary entries with user profile information from klaviyo.
                        Each entry must contain 'id' and 'email' as keys.
            list_id: The id of a list in klaviyo which all profiles will be added to.
                        This list_id is optional.

        Returns:
            True if all import jobs complete successfully
        """

        def format_profile(attributes):
            email = attributes.pop('$email')
            # in case of missing keys, set them to None. Klaviyo will ignore None values
            # and therefore the profile will not be updated with these values.
            city = attributes.pop('city', None)
            country = attributes.pop('country', None)
            zipcode = attributes.pop('zip_code', None)
            region = attributes.pop('region', None)
            profile = {
                "type": "profile",
                "attributes": {
                    "email": email,
                    "properties": {
                        k: v for k, v in attributes.items()
                    },
                    "location": {
                        'city': city,
                        'country': country,
                        'zip': zipcode,
                        'region': region}
                }
            }
            return profile

        successful = self._start_bulk_import_profile_job(
            format_method=format_profile,
            profiles=profiles,
            list_id=list_id
        )

        return successful

    def _start_bulk_import_profile_job(self, format_method: Callable[[Any], dict],
                                       profiles: List[Any], list_id: str = None) -> bool:
        """
        Chunks bulk jobs and posts to klaviyo
        Args:
            format_method: method to format the payload correctly
            profiles: list of profiles to be sent to klaviyo
            list_id: id of list to insert profiles into (optional)

        Returns:
            True if all jobs complete successfully

        """
        job_type = 'profile-bulk-import-job'
        url = self._KLAVIYO_BULK_UPSERT_PROFILES_URL

        # chunk list
        chunks = chunk_list(profiles, 1000)
        num_jobs = len(chunks)
        self.logger.info(f'Creating {num_jobs} jobs...')

        # process chunk in chunks
        for chunk in chunks:
            payload = self._get_payload_for_bulk_job(
                job_type=job_type,
                format_method=format_method,
                profiles=chunk,
                list_id=list_id
            )

            self._submit_bulk_import_profile_job(
                url=url,
                payload=payload
            )
            time.sleep(1)

        return True

    def _get_payload_for_bulk_job(self, job_type: str, format_method: Callable[[Any], dict],
                                  profiles: List[Any], list_id: str = None):
        """
        Helper method to get the correctly formatted payload for a bulk upload job.
        Args:
            job_type: job type (see klaviyo api docs)
            format_method: method to format the profile data correctly
            profiles: list of profiles to upload
            list_id: id of list to insert profiles into (optional)

        Returns:
            Formatted payload to upload using bulk job.
        """
        payload = {
            "data": {
                "type": job_type,
                "attributes": {
                    "profiles": {
                        "data": [format_method(profile) for profile in profiles]
                    }
                }
            }
        }

        if list_id:
            payload['data']['relationships'] = {
                "lists": {
                    "data": [
                        {
                            "type": "list",
                            "id": list_id
                        }
                    ]
                }
            }

        return payload

    def _submit_bulk_import_profile_job(self, url: str, payload: Dict) -> str:
        """
        Submit bulk import profile job
        Args:
            url: endpoint to post to
            payload: payload

        Returns:
            The job_id of the bulk import job created
        """
        response = self._request(method='POST', url=url, json=payload)
        response.raise_for_status()

        job_id = response.json()['data']['id']

        return job_id

    def get_list_members_count(self) -> tuple:
        """ Returns the number of members in a Klaviyo list.

        Returns:
            The nname of the list along with it's member count
        """
        url = self._KLAVIYO_LIST_MEMBERS_COUNT.replace('id', self.list_id)
        response = self._request(method='GET', url=url)
        profile_count = response.json()['data']['attributes']['profile_count']
        list_name = response.json()['data']['attributes']['name']
        self.logger.info(f'Got list with name {list_name} has {profile_count} members.')
        return list_name, profile_count

    def get_segment_members_count(self) -> tuple:
        """ Returns the number of members in a Klaviyo segment.

        Returns:
            The name of the segment along with it's member count
        """
        url = self._KLAVIYO_SEGMENT_MEMBERS_COUNT.replace('id', self.segment_id)
        response = self._request(method='GET', url=url)
        profile_count = response.json()['data']['attributes']['profile_count']
        segment_name = response.json()['data']['attributes']['name']
        self.logger.info(f'Got segment with name {segment_name} has {profile_count} members.')
        return segment_name, profile_count



from abc import ABC, abstractmethod
import logging
import requests
from requests.adapters import HTTPAdapter
from requests.structures import CaseInsensitiveDict
from time import sleep
from typing import Dict
from typing import Union
from urllib3 import Retry
from urllib.error import HTTPError
from sbcommons.utils import remove_email_info_from_text


class CrmClient(ABC):
    """ Base class for clients used to connect to CRM tools.

    Attributes:
        token (str, Dict[str, str]): Token needed to access the CRM's API. This is either a single
            API key or a Dict with two keys, the PUBLIC_API_KEY and PRIVATE_API_KEY.
        session (requests.Session): The session object for the Symplify connection.
        rate_limit (int): Amounts of second to wait after getting a rate limited error (HTTP 429),
            in order to make a successful request again.
        max_rate_limit_retries (int): Number of times to retry making a request if a rate limited
            error occurs (status code 429/Too Many Requests).
    """

    def __init__(self, token: Union[str, Dict[str, str]], rate_limit: int,
                 max_rate_limit_retries: int):
        assert rate_limit > 0
        assert max_rate_limit_retries > 0
        self.token = token
        self.max_rate_limit_retries = max_rate_limit_retries
        self.rate_limit = rate_limit
        self.session = requests.session()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, excl_value, exc_traceback):
        self.session.__exit__()

    @abstractmethod
    def connect(self):
        raise NotImplementedError("connect method should be implemented by children "
                                  "classes of CrmClient.")

    @abstractmethod
    def _build_base_header(self) -> CaseInsensitiveDict:
        """ Method that should return the session header content as a dictionary. """
        raise NotImplementedError("_build_base_header method should be implemented by children "
                                  "classes of CrmClient.")

    @property
    @abstractmethod
    def logger(self) -> logging.Logger:
        """ Returns the logger object to be used by the instance. This must be implemented by
            child classes, using a distinct logger name corresponding to the child class. """
        raise NotImplementedError("Classes inheriting from CrmClient should implement the "
                                  "logger method with a @property decorator.")

    @staticmethod
    def _retry_adapter(retries=5,
                       backoff_factor=1.0,
                       status_forcelist=(500, 501, 502, 503, 504)):
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        return HTTPAdapter(max_retries=retry)

    def _request(self, method: str = 'GET', url: str = None, verbose: bool = False, *args,
                 **kwargs) -> requests.Response:
        """ Helper/Wrapper method around the sessions.request API.

        Args:
            method: The type of request.
            url: The url to use for the request.
            verbose: Whether to write out the response of the call as text with the logger.

        Returns:
            A requests.Response object corresponding to the request's response.

        Raises:
            HTTPError if the response status codeis different than 200 and 429, or if it is 429 but
                all attempts to retry the request fail.
        """
        for i in range(self.max_rate_limit_retries):
            response = self.session.request(method, url, *args, **kwargs)

            # Continue the loop only if we got a rate limit error
            if response.status_code != 429:
                break

            # Rate limited error (we should make one request per minute).
            self.logger.warning(
                f'Rate limited by the CRM server API. Sleeping for {self.rate_limit} seconds'
            )
            # Wait and re-try the request
            sleep(self.rate_limit)

        # If an error occurred during the request, raise an exception
        if response.status_code not in [200,201,202]:
            response_content_clean = remove_email_info_from_text(str(response.content))
            e_msg = f'HTTP response code: {response.status_code} - Response reason: ' \
                f'{response.reason} and content {response_content_clean}'
            self.logger.error(e_msg)
            raise HTTPError(url, response.status_code, e_msg, None, None)

        if verbose:
            self.logger.info(f"Response text: {response.text}")
        return response

    def post_list_sync(self, list_data: Union[bytes, Dict], import_type: str = 'ADD') -> bool:
        """ Posts a list of customers to the CRM tool - synchronous call."""
        raise NotImplementedError(f'post_list_sync has not been implemented in {self.__class__}')

    def post_survey_candidates(self, **kwargs):
        """ Wrapper around post_list_sync. Used to achieve polymorphism with other CrmClient
        subclasses such as KlaviyoClient. We do this because we might use different methods to post
        survey candidates in different CRM systems. """
        return self.post_list_sync(**kwargs)



import logging
import sys


def get_logger(logger_name: str, log_level: int = logging.INFO) -> logging.Logger:
    logger = logging.getLogger(logger_name)
    h = logging.StreamHandler(sys.stdout)
    log_format = '[%(levelname)s] %(asctime)s - %(name)s: %(message)s'
    h.setFormatter(logging.Formatter(log_format))
    logger.addHandler(h)
    logger.setLevel(log_level)
    logger.propagate = False

    return logger



from sbcommons.logging import lambda_logger
import requests

logger = lambda_logger.get_logger(__name__)


def post_to_webhook(service: str, webhook_url: str, json: dict) -> bool:
    try:
        response = requests.post(webhook_url, json=json)
        if not response.status_code == 200:
            logger.error(
                f'Tried to post message on {service} webhook but failed, status code: {response.status_code}'
            )
            return False
    except Exception as e:
        logger.error(f'Exception when posting to {service}, exception args: {e.args}')
        return False

    return True



from abc import ABC, abstractmethod
from datetime import datetime
import enum
from typing import Any
from typing import List
from typing import Iterable


class DataType(enum.Enum):
    STRING = str
    TIMESTAMP = datetime


class TranslationEntry:
    def __init__(self, db_column: str,
                 crm_attr_name: str,
                 data_type: DataType,
                 timestamp_format: str = None):
        self.db_column = db_column
        self.crm_attr_name = crm_attr_name
        self.data_type = data_type
        if data_type.value == datetime:
            assert timestamp_format is not None
            self.timestamp_format = timestamp_format


class CrmParser(ABC):
    """
    Attributes:
        translation_dict: A dictionary that maps a database column name to a TranslationEntry,
            which is used to rename the column to the right name for the CRM client and cast its
            values to the right data type.
    """
    def __init__(self, translation_entries: List[TranslationEntry]):
        self.translation_dict = {entry.db_column: entry for entry in translation_entries}

    def parse_header(self, header: List[str]) -> List[str]:
        """ Parses header, translating each column name to the right CRM attribute name """
        return [self.translation_dict[x].crm_attr_name for x in header]

    @abstractmethod
    def parse(self, header: List[str], rows: List[Iterable[Any]]):
        pass

    def _sanitize(self, db_column_name: str, row_val: Any, delimiter: str = '') -> str:
        data_type = self.translation_dict[db_column_name].data_type
        if row_val:
            if data_type.value == str:
                row_val = str(row_val).replace(delimiter, '')
            elif data_type.value == datetime:
                ts_format = self.translation_dict[db_column_name].timestamp_format
                row_val = row_val.strftime(ts_format)
        elif row_val is None:
            row_val = str(row_val).replace('None', '')

        return str(row_val)



from typing import Any
from typing import List
from typing import Iterable

from sbcommons.crm.parser import CrmParser
from sbcommons.crm.parser import TranslationEntry


class SymplifyParser(CrmParser):
    """ Class for parsing query results into a format compatible with the SymplifyClient.

    Attributes:
        translation_dict: A dictionary that maps a database column name to a TranslationEntry,
            which is used to rename the column to the right name for Symplify and cast its
            values to the right data type. Inherited from Parser.
        delimiter: What delimiter to use to separate fields when parsing the input.
    """
    def __init__(self, translation_entries: List[TranslationEntry], delimiter: str = '|'):
        CrmParser.__init__(self, translation_entries)
        self.delimiter = delimiter

    def source_headers(self) -> List[str]:
        """ Returns the column names as expected from the query results. """
        return [db_column_name for db_column_name in self.translation_dict]

    def parse(self, header: List[str], rows: List[Iterable[Any]]) -> bytes:
        """ Converts the <header> and <rows> to a format compatible with SymplifyClient.

        Args:
            header: The column names for each field in <rows>.
            rows: A list of iterables where each iterable has the values specified by the <header>.

        Returns:
            A byte sequence that can be used in SymplifyClient calls such as a post_list_sync call.
        """
        parsed_header = self.parse_header(header)
        header_str = self.delimiter.join(parsed_header)
        plain_text_data = '\n'.join([header_str] + [self._parse_row(header, row) for row in rows])
        return plain_text_data.encode('utf-8')

    def _parse_row(self, header: List[str], row: Iterable[Any]):
        return f'{self.delimiter}'.join([self._sanitize(db_column_name, val, self.delimiter)
                                         for db_column_name, val in zip(header, row)])



from sbcommons.messaging import webhooks


def send_message(webhook_url: str, title: str, msg: str) -> bool:
    data = {
        "blocks": [
            {
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": title
                }
            },
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": msg
                }
            }
        ]
    }

    webhooks.post_to_webhook(service='slack', webhook_url=webhook_url, json=data)
